
/*
 *  Test driver for 16550A UART
 *
 *  based on linux/drivers/char/serial.c
 *
 */

#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
	#include <linux/modversions.h>
	#define MODVERSIONS
#endif

#define __NO_VERSION__		/* version independence			*/
#include <linux/version.h>	/* not included due to __NO_VERSION__	*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>	        /* contains kmalloc, memset, kfree etc.	*/
#include <linux/errno.h>	/* Error codes				*/
#include <linux/types.h>	/* contains size_t etc.			*/
#include <linux/ioctl.h>	/* for IOCTL things			*/
#include <asm/uaccess.h> 	/* has put_user fxn 			*/
//#include <asm/system.h>
//#include <linux/config.h>
#include <linux/serial.h>
#include <linux/serialP.h>
//#include <linux/serial_reg.h>
//#include <asm/serial.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
//#include <linux/tty_flip.h>
//#include <linux/init.h>
#include <asm/io.h>
#include "local_serial.h"

#define _INLINE_ inline

static char *serial_name = "Mandeep's Serial driver";

#define DRIVER_VERSION "v1.0"
#define DRIVER_AUTHOR "Mandeep Sandhu"
#define DRIVER_DESC "Test Serial Driver"

/* Module Info*/
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

#define TRANSMIT_CHAR	't'

#define TEST_MAJOR	4	/* Serial port's MAJOR	*/
#define TEST_MINOR	64	/* ttyS0's minor number	*/

#define SERIAL_NAME	"ttytest"

#define NR_PORTS	1	/* only one port supported */

static struct my_struct	*gbl_info;

static DECLARE_TASK_QUEUE(tq_serial);

static struct tty_driver sd_16550A;
static int serial_refcount;

static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

static void do_serial_bh(void)
{
	run_task_queue(&tq_serial);
}

void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;

	do {
		i = inb(MYPORT_16550A + 5);
		if(i & 1)
		   printk(KERN_ALERT "received char>>%c\n",inb(MYPORT_16550A));
	} while(i & 1);
	/*
	gbl_info->xmit.buf[gbl_info->xmit.tail] = i;

	gbl_info->xmit.tail = (gbl_info->xmit.tail + 1)
				& (gbl_info->xmit_fifo_size-1);
	*/
	mark_bh(SERIAL_BH);
}

static int get_serial_info(struct my_struct *info,
			   struct serial_struct *retinfo)
{
	struct serial_struct tmp;
	//struct serial_state *state = info->state;

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->tty->driver.type;
	//tmp.line = info->tty->line;
	tmp.port = info->port;
	tmp.port_high = 0;
	tmp.irq = 4;//info->tty->irq;
	tmp.flags = info->flags;
	tmp.xmit_fifo_size = info->xmit_fifo_size;
	tmp.baud_base = 9600;//info->tty->baud_base;
	/*tmp.close_delay = info->tty->close_delay;
	tmp.closing_wait = info->tty->closing_wait;
	tmp.custom_divisor = info->tty->custom_divisor;*/
	//tmp.hub6 = info->tty->hub6;
	tmp.io_type = info->io_type;
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int mss_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	struct my_struct *info = (struct my_struct *)tty->driver_data;

	switch (cmd)
	{
		case TIOCGSERIAL:
			return get_serial_info(info,
					       (struct serial_struct *) arg);
		default:
			return -ENOIOCTLCMD;
	}
}

static int startup(struct my_struct *info)
{
	unsigned long 	flags;
	int		retval = 0;
	unsigned long page;

	save_flags(flags); cli();

	if(!info)
		goto errout;

	page = get_zeroed_page(GFP_KERNEL);
	if (!page)
		return -ENOMEM;

	if (info->xmit.buf)
		free_page(page);
	else
		info->xmit.buf = (unsigned char *) page;

	//Register the INTERRUPT HANDLER
	retval = request_irq(4, irq_handler, SA_SHIRQ,"my_test_handler", info);

	if(retval)
	{
		printk(KERN_ALERT "Unable to register IRQ handler(err - %d)", retval);
		goto errout;
	}

	/*
	 * Now, initialize the UART
	 */
	outb(0, MYPORT_16550A + 1);	/* Turn off interrupts - Port 1 */
	outb(0, MYPORT_16550A + 1);	/* Disable interrupts - bits 0-> 0 */
	outb(0x80, MYPORT_16550A + 3);	/* enable DLAB bit 7 -> 1 */
	//Set baud rate to 9600 = (Max freq(115.2 MHz)/12)
	//12 stored in 16 bit reg. LSB=0x0C, MSB=0x00
	outb(0x0C, MYPORT_16550A + 0);	/* Set Divisor LSB */
	outb(0x00, MYPORT_16550A + 1);	/* Set Divisor MSB */
	outb(0x03, MYPORT_16550A + 3);	/* 8 bits, No parity, 1 Stop bit */
	//outb(0xC7, MYPORT_16550A + 2);	/* Enable FIFO (16550A) */
	outb(0x0B, MYPORT_16550A + 4);	/* Turn on DTR, RTS and OUT2 */
	outb(0x01, MYPORT_16550A + 1);	/* Interrupt when data received */

	if(!info)
		return 0;

	info->xmit.head = info->xmit.tail = 0;

	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
	//printk(KERN_ALERT "startup done!\n");
	return 0;

errout:
	restore_flags(flags);
	return retval;
}

static int mss_write_room(struct tty_struct *tty)
{
	struct my_struct *info = (struct my_struct *)tty->driver_data;

	return CIRC_SPACE(info->xmit.head, info->xmit.tail, SERIAL_XMIT_SIZE);
}

static int mss_write(struct tty_struct * tty, int from_user,
		    const unsigned char *buf, int count)
{
	unsigned char c = TRANSMIT_CHAR;
	printk(KERN_ALERT "In write!\n");
	outb(c,MYPORT_16550A);	//Transmit 't' all the time!
	return count;
}

static void mss_close(struct tty_struct *tty, struct file * filp)
{
	//printk(KERN_ALERT "In CLOSE!\n");
	MOD_DEC_USE_COUNT;
}

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its my_struct
 * structure into the IRQ chain. It also performs the serial-specific
 * initialization for the tty structure.
 */
static int mss_open(struct tty_struct *tty, struct file * filp)
{
	struct my_struct	*info;
	int 			retval;
	//unsigned long		page;

	printk(KERN_ALERT "In OPEN!\n");

	MOD_INC_USE_COUNT;

	info = (struct my_struct *)kmalloc(sizeof(struct my_struct), GFP_KERNEL);
	memset(info, 0, sizeof(struct my_struct));

	info->magic = 0;
	info->port = MYPORT_16550A;
	info->flags = 0;
	info->io_type = 0;
	//info->io_mem_base = 0;
	info->xmit_fifo_size = SERIAL_XMIT_SIZE;
	//info->tqueue.routine = do_softint;
	info->tqueue.data = info;

	tty->driver_data = info;
	info->tty = tty;
	gbl_info = info;

	/*
	 * Start up serial port
	 */
	retval = startup(info);
	if (retval)
	{
		MOD_DEC_USE_COUNT;
		return retval;
	}

	return 0;//success!
}

/*
 * The serial driver load-time initialization code!
 */
static int __init mss_init(void)
{
	int result;

	//free_irq(4,NULL);
	init_bh(SERIAL_BH, do_serial_bh);

	memset(&sd_16550A, 0, sizeof(struct tty_driver));
	sd_16550A.magic = TTY_DRIVER_MAGIC;
	sd_16550A.driver_name = SERIAL_NAME;
	sd_16550A.name = "ttyS";
	sd_16550A.major = TTY_MAJOR;
	sd_16550A.minor_start = TEST_MINOR;
	sd_16550A.name_base = 0;
	sd_16550A.num = NR_PORTS;
	sd_16550A.type = TTY_DRIVER_TYPE_SERIAL;
	sd_16550A.subtype = SERIAL_TYPE_NORMAL;
	sd_16550A.init_termios = tty_std_termios;
	sd_16550A.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	sd_16550A.flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
	sd_16550A.refcount = &serial_refcount;
	sd_16550A.table = serial_table;
	sd_16550A.termios = serial_termios;
	sd_16550A.termios_locked = serial_termios_locked;

	sd_16550A.open = mss_open;
	sd_16550A.close = mss_close;
	sd_16550A.write = mss_write;
	sd_16550A.write_room = mss_write_room;
	sd_16550A.ioctl = mss_ioctl;

	sd_16550A.next = NULL;
	sd_16550A.prev = NULL;

	result = tty_register_driver(&sd_16550A);

	if(result)
	{
		panic("Couldn't register serial driver\n");
		return result;
	}

	printk(KERN_ALERT "Initialization succesfull!\n");
	return 0;
}

static void __exit mss_exit(void)
{
	unsigned long flags;

	/* printk("Unloading %s: version %s\n", serial_name, serial_version); */
	save_flags(flags); cli();
        remove_bh(SERIAL_BH);

	if (tty_unregister_driver(&sd_16550A))
		printk("serial: failed to unregister serial driver\n");

	restore_flags(flags);
	free_irq(4,gbl_info);
	return;
}

module_init(mss_init);
module_exit(mss_exit);
